#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import random
from typing import Any, Dict, List, Optional, Union

import pytest
import pytest_asyncio
from argilla_server.enums import MetadataPropertyType, QuestionType, ResponseStatusFilter, SimilarityOrder
from argilla_server.models import Dataset, Question, Record, User, VectorSettings
from argilla_server.search_engine import (
    FloatMetadataFilter,
    IntegerMetadataFilter,
    SortBy,
    SuggestionFilterScope,
    TermsFilter,
    TermsMetadataFilter,
    TextQuery,
    UserResponseStatusFilter,
)
from argilla_server.search_engine.commons import (
    ALL_RESPONSES_STATUSES_FIELD,
    BaseElasticAndOpenSearchEngine,
    es_index_name_for_dataset,
)
from argilla_server.settings import settings as server_settings
from opensearchpy import OpenSearch
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Session

from tests.factories import (
    DatasetFactory,
    FloatMetadataPropertyFactory,
    IntegerMetadataPropertyFactory,
    LabelSelectionQuestionFactory,
    MultiLabelSelectionQuestionFactory,
    QuestionFactory,
    RatingQuestionFactory,
    RecordFactory,
    ResponseFactory,
    SuggestionFactory,
    TermsMetadataPropertyFactory,
    TextFieldFactory,
    TextQuestionFactory,
    UserFactory,
    VectorFactory,
    VectorSettingsFactory,
)


@pytest_asyncio.fixture(scope="function")
async def dataset_for_pagination(opensearch: OpenSearch):
    from opensearchpy import helpers

    dataset = await DatasetFactory.create(fields=[], questions=[])
    records = await RecordFactory.create_batch(size=100, dataset=dataset)
    await dataset.awaitable_attrs.records
    index_name = es_index_name_for_dataset(dataset)

    opensearch.indices.create(index=index_name, body={"mappings": {"properties": {"id": {"type": "keyword"}}}})

    bulk_actions = [
        {
            "_op_type": "index",
            "_id": record.id,
            "_index": index_name,
            "id": record.id,
            "fields": {"text": "The same text for all documents"},
        }
        for record in records
    ]

    helpers.bulk(client=opensearch, actions=bulk_actions)

    opensearch.indices.refresh(index=index_name)

    yield dataset

    opensearch.indices.delete(index=index_name)


@pytest_asyncio.fixture(scope="function")
@pytest.mark.asyncio
async def test_banking_sentiment_dataset(
    search_engine: BaseElasticAndOpenSearchEngine, opensearch: OpenSearch
) -> Dataset:
    text_question = await TextQuestionFactory()
    rating_question = await RatingQuestionFactory()

    dataset = await DatasetFactory.create(
        fields=[
            await TextFieldFactory.create(name="textId"),
            await TextFieldFactory.create(name="text"),
            await TextFieldFactory.create(name="label"),
        ],
        metadata_properties=[
            await TermsMetadataPropertyFactory.create(name="label"),
            await IntegerMetadataPropertyFactory.create(name="textId"),
            await FloatMetadataPropertyFactory.create(name="seq_float"),
        ],
        questions=[text_question, rating_question],
    )

    records = [
        await RecordFactory.create(
            dataset=dataset,
            fields={"textId": "00000", "text": "My card payment had the wrong exchange rate", "label": "negative"},
            metadata_={"label": None, "textId": "00000", "seq_float": 0.13},
            responses=[],
        ),
        await RecordFactory.create(
            dataset=dataset,
            fields={
                "textId": "00001",
                "text": "I believe that a card payment I made was cancelled.",
                "label": "neutral",
            },
            metadata_={"label": "neutral", "textId": "00001", "seq_float": 0.343},
            responses=[],
        ),
        await RecordFactory.create(
            dataset=dataset,
            fields={
                "textId": "00002",
                "text": "I tried to make a payment with my card and it was declined.",
                "label": "negative",
            },
            metadata_={"label": "negative", "textId": "00002", "seq_float": 1.123},
            responses=[],
        ),
        await RecordFactory.create(
            dataset=dataset,
            fields={
                "textId": "00003",
                "text": "My credit card was declined when I tried to make a payment.",
                "label": "negative",
            },
            metadata_={"label": "negative", "textId": "00003", "seq_float": 12.13},
            responses=[],
        ),
        await RecordFactory.create(
            dataset=dataset,
            fields={
                "textId": "00004",
                "text": "I made a successful payment towards my mortgage loan earlier today.",
                "label": "positive",
            },
            metadata_={"label": "positive", "textId": "00004", "seq_float": -120.13},
            responses=[],
        ),
        await RecordFactory.create(
            dataset=dataset,
            fields={
                "textId": "00005",
                "text": "Please confirm the receipt of my payment for the credit card bill due on the 15th.",
                "label": "neutral",
            },
            metadata_={"label": "neutral", "textId": "00005", "seq_float": -149.13},
            responses=[],
        ),
        await RecordFactory.create(
            dataset=dataset,
            fields={
                "textId": "00006",
                "text": "Why was I charged for getting cash?",
                "label": "neutral",
            },
            metadata_={"label": "neutral", "textId": "00006", "seq_float": 200.13},
            responses=[],
        ),
        await RecordFactory.create(
            dataset=dataset,
            fields={
                "textId": "00007",
                "text": "Why was I charged for getting cash?",
                "label": "neutral",
            },
            metadata_={"label": "neutral", "textId": "00007", "seq_float": 20020.13},
            responses=[],
        ),
        await RecordFactory.create(
            dataset=dataset,
            fields={
                "textId": "00008",
                "text": "I deposited cash into my account a week ago and it is still not available,"
                " please tell me why? I need the cash back now.",
                "label": "negative",
            },
            metadata_={"label": "negative", "textId": "00008", "seq_float": 120020.13},
            responses=[],
        ),
    ]

    await refresh_dataset(dataset)
    await refresh_records(records)
    await dataset.awaitable_attrs.records

    await search_engine.create_index(dataset)
    await search_engine.index_records(dataset, records=records)

    return dataset


@pytest_asyncio.fixture(scope="function")
@pytest.mark.asyncio
async def test_banking_sentiment_dataset_with_vectors(
    search_engine: BaseElasticAndOpenSearchEngine, opensearch: OpenSearch, test_banking_sentiment_dataset: Dataset
) -> Dataset:
    vectors_settings = await VectorSettingsFactory.create_batch(5, dataset=test_banking_sentiment_dataset)

    for settings in vectors_settings:
        await search_engine.configure_index_vectors(settings)

    vectors = []
    for record in test_banking_sentiment_dataset.records:
        for settings in vectors_settings:
            vectors.append(
                await VectorFactory.create(
                    vector_settings=settings,
                    record=record,
                    value=[random.uniform(-10, 10) for _ in range(0, settings.dimensions)],
                )
            )
        await record.awaitable_attrs.vectors

    await search_engine.set_records_vectors(test_banking_sentiment_dataset, vectors=vectors)

    await refresh_dataset(test_banking_sentiment_dataset)
    await test_banking_sentiment_dataset.awaitable_attrs.records

    return test_banking_sentiment_dataset


async def refresh_dataset(dataset: Dataset):
    await dataset.awaitable_attrs.fields
    await dataset.awaitable_attrs.questions
    await dataset.awaitable_attrs.metadata_properties
    await dataset.awaitable_attrs.vectors_settings


async def refresh_records(records: List[Record]):
    for record in records:
        await record.awaitable_attrs.suggestions
        await record.awaitable_attrs.responses
        await record.awaitable_attrs.vectors


def _expected_value_for_question(question: Question) -> Dict[str, Any]:
    if question.type in [QuestionType.label_selection, QuestionType.multi_label_selection]:
        return {"type": "keyword"}
    elif question.type == QuestionType.text:
        return {"type": "text", "index": False}
    elif question.type == QuestionType.rating:
        return {"type": "integer"}
    elif question.type == QuestionType.ranking:
        return {"type": "nested"}
    else:
        return {}


@pytest.mark.asyncio
@pytest.mark.skipif(
    not server_settings.search_engine in ["elasticsearch", "opensearch"],
    reason="Running on elasticsearch/opensearch engine",
)
class TestBaseElasticAndOpenSearchEngine:
    """
    This test suite check all common methods for ElasticSearch and OpenSearch engines.

    If you need to test a specific method for one of these engines, you should the
    `test_elasticsearch.py` and `test_opensearch.py` files.

    """

    async def test_create_index_for_dataset(
        self, search_engine: BaseElasticAndOpenSearchEngine, db: "AsyncSession", opensearch: OpenSearch
    ):
        dataset = await DatasetFactory.create()

        await refresh_dataset(dataset)

        await search_engine.create_index(dataset)

        index_name = es_index_name_for_dataset(dataset)
        assert opensearch.indices.exists(index=index_name)

        index = opensearch.indices.get(index=index_name)[index_name]
        assert index["mappings"] == {
            "dynamic": "strict",
            "dynamic_templates": [
                {
                    "status_responses": {
                        "mapping": {"type": "keyword", "copy_to": ALL_RESPONSES_STATUSES_FIELD},
                        "path_match": "responses.*.status",
                    }
                }
            ],
            "properties": {
                "id": {"type": "keyword"},
                "inserted_at": {"type": "date_nanos"},
                "updated_at": {"type": "date_nanos"},
                ALL_RESPONSES_STATUSES_FIELD: {"type": "keyword"},
                "responses": {"dynamic": "true", "type": "object"},
                "metadata": {"dynamic": "false", "type": "object"},
            },
        }

        assert index["settings"]["index"]["number_of_shards"] == str(search_engine.number_of_shards)
        assert index["settings"]["index"]["number_of_replicas"] == str(search_engine.number_of_replicas)

    async def test_create_index_for_dataset_with_fields(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        db: Session,
    ):
        text_fields = await TextFieldFactory.create_batch(5)
        dataset = await DatasetFactory.create(fields=text_fields)

        await refresh_dataset(dataset)

        await search_engine.create_index(dataset)

        index_name = es_index_name_for_dataset(dataset)
        assert opensearch.indices.exists(index=index_name)

        index = opensearch.indices.get(index=index_name)[index_name]
        assert index["mappings"] == {
            "dynamic": "strict",
            "dynamic_templates": [
                {
                    "status_responses": {
                        "mapping": {"type": "keyword", "copy_to": ALL_RESPONSES_STATUSES_FIELD},
                        "path_match": "responses.*.status",
                    }
                }
            ],
            "properties": {
                "id": {"type": "keyword"},
                "inserted_at": {"type": "date_nanos"},
                "updated_at": {"type": "date_nanos"},
                ALL_RESPONSES_STATUSES_FIELD: {"type": "keyword"},
                "fields": {"properties": {field.name: {"type": "text"} for field in dataset.fields}},
                "responses": {"type": "object", "dynamic": "true"},
                "metadata": {"dynamic": "false", "type": "object"},
            },
        }

    async def test_create_metadata_property(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        db: "AsyncSession",
    ):
        text_field = await TextFieldFactory.create(name="field")

        terms_property = await TermsMetadataPropertyFactory.create(name="terms")
        integer_property = await IntegerMetadataPropertyFactory.create(name="integer")
        float_property = await FloatMetadataPropertyFactory.create(name="float")

        metadata_properties = [terms_property, integer_property]

        dataset = await DatasetFactory.create(fields=[text_field], metadata_properties=metadata_properties)

        await refresh_dataset(dataset)

        await search_engine.create_index(dataset)
        await search_engine.configure_metadata_property(dataset, float_property)

        index_name = es_index_name_for_dataset(dataset)
        assert opensearch.indices.exists(index=index_name)

        index = opensearch.indices.get(index=index_name)[index_name]
        assert index["mappings"]["properties"]["metadata"]["properties"][str(float_property.name)] == {"type": "float"}

    async def test_create_index_for_dataset_with_metadata_properties(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        db: "AsyncSession",
    ):
        text_field = await TextFieldFactory.create(name="field")

        terms_property = await TermsMetadataPropertyFactory.create(name="terms")
        integer_property = await IntegerMetadataPropertyFactory.create(name="integer")
        float_property = await FloatMetadataPropertyFactory.create(name="float")

        metadata_properties = [terms_property, integer_property, float_property]

        dataset = await DatasetFactory.create(fields=[text_field], metadata_properties=metadata_properties)

        await refresh_dataset(dataset)

        await search_engine.create_index(dataset)

        index_name = es_index_name_for_dataset(dataset)
        assert opensearch.indices.exists(index=index_name)

        index = opensearch.indices.get(index=index_name)[index_name]
        assert index["mappings"] == {
            "dynamic": "strict",
            "dynamic_templates": [
                {
                    "status_responses": {
                        "mapping": {"type": "keyword", "copy_to": ALL_RESPONSES_STATUSES_FIELD},
                        "path_match": "responses.*.status",
                    }
                }
            ],
            "properties": {
                "id": {"type": "keyword"},
                "inserted_at": {"type": "date_nanos"},
                "updated_at": {"type": "date_nanos"},
                ALL_RESPONSES_STATUSES_FIELD: {"type": "keyword"},
                "fields": {"properties": {"field": {"type": "text"}}},
                "responses": {"type": "object", "dynamic": "true"},
                "metadata": {
                    "dynamic": "false",
                    "properties": {
                        str(terms_property.name): {"type": "keyword"},
                        str(integer_property.name): {"type": "long"},
                        str(float_property.name): {"type": "float"},
                    },
                },
            },
        }

    @pytest.mark.parametrize(
        argnames=("text_ann_size", "rating_ann_size"),
        argvalues=[(random.randint(1, 9), random.randint(1, 9)) for _ in range(1, 5)],
    )
    async def test_create_index_for_dataset_with_questions(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        text_ann_size: int,
        rating_ann_size: int,
    ):
        text_questions = await TextQuestionFactory.create_batch(size=text_ann_size)
        rating_questions = await RatingQuestionFactory.create_batch(size=rating_ann_size)
        label_questions = await LabelSelectionQuestionFactory.create_batch(size=text_ann_size)
        multilabel_questions = await MultiLabelSelectionQuestionFactory.create_batch(size=rating_ann_size)

        all_questions = text_questions + rating_questions + label_questions + multilabel_questions
        dataset = await DatasetFactory.create(questions=all_questions)

        await refresh_dataset(dataset)

        await search_engine.create_index(dataset)

        index_name = es_index_name_for_dataset(dataset)
        assert opensearch.indices.exists(index=index_name)

        index = opensearch.indices.get(index=index_name)[index_name]
        assert index["mappings"] == {
            "dynamic": "strict",
            "properties": {
                "id": {"type": "keyword"},
                "inserted_at": {"type": "date_nanos"},
                "updated_at": {"type": "date_nanos"},
                ALL_RESPONSES_STATUSES_FIELD: {"type": "keyword"},
                "responses": {"dynamic": "true", "type": "object"},
                "metadata": {"dynamic": "false", "type": "object"},
                "suggestions": {
                    "properties": {
                        question.name: {
                            "properties": {
                                "type": {"type": "keyword"},
                                "agent": {"type": "keyword"},
                                "value": _expected_value_for_question(question),
                                "score": {"type": "float"},
                            }
                        }
                        for question in all_questions
                    }
                },
            },
            "dynamic_templates": [
                {
                    "status_responses": {
                        "mapping": {"type": "keyword", "copy_to": ALL_RESPONSES_STATUSES_FIELD},
                        "path_match": "responses.*.status",
                    }
                },
                *[
                    config
                    for question in all_questions
                    for config in [
                        {
                            f"{question.name}_responses": {
                                "mapping": _expected_value_for_question(question),
                                "path_match": f"responses.*.values.{question.name}",
                            }
                        },
                    ]
                ],
            ],
        }

    @pytest.mark.parametrize(
        ("query", "expected_items"),
        [
            ("card", 5),
            ("account", 1),
            ("payment", 6),
            ("cash", 3),
            ("negative", 4),
            ("00000", 1),
            ("card payment", 5),
            ("nothing", 0),
            (TextQuery(q="card"), 5),
            (TextQuery(q="account"), 1),
            (TextQuery(q="payment"), 6),
            (TextQuery(q="cash"), 3),
            (TextQuery(q="card payment"), 5),
            (TextQuery(q="nothing"), 0),
            (TextQuery(q="rate negative"), 1),  # Terms are found in two different fields
            (TextQuery(q="negative", field="label"), 4),
            (TextQuery(q="00000", field="textId"), 1),
            (TextQuery(q="card payment", field="text"), 5),
        ],
    )
    async def test_search_with_query_string(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        test_banking_sentiment_dataset: Dataset,
        query: Union[str, TextQuery],
        expected_items: int,
    ):
        result = await search_engine.search(test_banking_sentiment_dataset, query=query)

        assert len(result.items) == expected_items
        assert result.total == expected_items

        scores = [item.score > 0 for item in result.items]
        assert all(map(lambda s: s > 0, scores))

        sorted_scores = scores.copy()
        sorted_scores.sort(reverse=True)

        assert scores == sorted_scores

    @pytest.mark.parametrize(
        "statuses, expected_items",
        [
            ([], 6),
            ([ResponseStatusFilter.missing], 6),
            ([ResponseStatusFilter.draft], 2),
            ([ResponseStatusFilter.submitted], 2),
            ([ResponseStatusFilter.discarded], 2),
            ([ResponseStatusFilter.missing, ResponseStatusFilter.draft], 6),
            ([ResponseStatusFilter.submitted, ResponseStatusFilter.discarded], 4),
            ([ResponseStatusFilter.missing, ResponseStatusFilter.draft, ResponseStatusFilter.discarded], 6),
        ],
    )
    async def test_search_with_response_status_filter(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        test_banking_sentiment_dataset: Dataset,
        statuses: List[ResponseStatusFilter],
        expected_items: int,
    ):
        user = await UserFactory.create()

        await self._configure_record_responses(
            opensearch, test_banking_sentiment_dataset, statuses, expected_items, user
        )

        result = await search_engine.search(
            test_banking_sentiment_dataset,
            query=TextQuery(q="payment"),
            user_response_status_filter=UserResponseStatusFilter(user=user, statuses=statuses),
        )
        assert len(result.items) == expected_items
        assert result.total == expected_items

    @pytest.mark.parametrize(
        "statuses, expected_items",
        [
            ([], 9),
            ([ResponseStatusFilter.missing], 3),
            ([ResponseStatusFilter.draft], 5),
            ([ResponseStatusFilter.submitted], 3),
            ([ResponseStatusFilter.discarded], 2),
            ([ResponseStatusFilter.missing, ResponseStatusFilter.draft], 5),
            ([ResponseStatusFilter.submitted, ResponseStatusFilter.discarded], 3),
            ([ResponseStatusFilter.missing, ResponseStatusFilter.draft, ResponseStatusFilter.discarded], 4),
        ],
    )
    async def test_search_with_response_status_filter_with_no_user(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        test_banking_sentiment_dataset: Dataset,
        statuses: List[ResponseStatusFilter],
        expected_items: int,
    ):
        await self._configure_record_responses(opensearch, test_banking_sentiment_dataset, statuses, expected_items)

        result = await search_engine.search(
            test_banking_sentiment_dataset,
            user_response_status_filter=UserResponseStatusFilter(statuses=statuses, user=None),
        )

        assert len(result.items) == expected_items
        assert result.total == expected_items

    @pytest.mark.parametrize(
        ("metadata_filters_config", "expected_items"),
        [
            ([{"name": "label", "values": ["neutral"]}], 4),
            ([{"name": "label", "values": ["positive"]}], 1),
            ([{"name": "label", "values": ["neutral", "positive"]}], 5),
            ([{"name": "textId", "ge": 3, "le": 4}], 2),
            ([{"name": "textId", "ge": 3, "le": 3}], 1),
            ([{"name": "textId", "ge": 3}], 6),
            ([{"name": "textId", "le": 4}], 5),
            ([{"name": "seq_float", "ge": 0.0, "le": 12.03}], 3),
            ([{"name": "seq_float", "ge": 0.13, "le": 0.13}], 1),
            ([{"name": "seq_float", "ge": 0.0}], 7),
            ([{"name": "seq_float", "le": 12.03}], 5),
        ],
    )
    async def test_search_with_metadata_filter(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        test_banking_sentiment_dataset: Dataset,
        metadata_filters_config: List[dict],
        expected_items: int,
    ):
        metadata_filters = []
        for metadata_filter_config in metadata_filters_config:
            name = metadata_filter_config.pop("name")
            for metadata_property in test_banking_sentiment_dataset.metadata_properties:
                if name == metadata_property.name:
                    if metadata_property.type == MetadataPropertyType.terms:
                        filter_cls = TermsMetadataFilter
                    elif metadata_property.type == MetadataPropertyType.integer:
                        filter_cls = IntegerMetadataFilter
                    else:
                        filter_cls = FloatMetadataFilter
                    metadata_filters.append(filter_cls(metadata_property=metadata_property, **metadata_filter_config))
                    break

        result = await search_engine.search(test_banking_sentiment_dataset, metadata_filters=metadata_filters)
        assert len(result.items) == expected_items
        assert result.total == expected_items

    async def test_search_with_no_query(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        test_banking_sentiment_dataset: Dataset,
    ):
        result = await search_engine.search(test_banking_sentiment_dataset)
        assert len(result.items) == len(test_banking_sentiment_dataset.records)
        assert result.total == len(test_banking_sentiment_dataset.records)

        result_scores = set([item.score for item in result.items])
        assert result_scores == {1.0}

    async def test_search_with_response_status_filter_does_not_affect_the_result_scores(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        test_banking_sentiment_dataset: Dataset,
    ):
        user = await UserFactory.create()

        all_statuses = [ResponseStatusFilter.missing, ResponseStatusFilter.draft, ResponseStatusFilter.discarded]
        await self._configure_record_responses(
            opensearch, test_banking_sentiment_dataset, all_statuses, len(test_banking_sentiment_dataset.records), user
        )

        no_filter_results = await search_engine.search(test_banking_sentiment_dataset, query=TextQuery(q="payment"))

        results = await search_engine.search(
            test_banking_sentiment_dataset,
            query=TextQuery(q="payment"),
            user_response_status_filter=UserResponseStatusFilter(user=user, statuses=all_statuses),
        )

        assert len(no_filter_results.items) == len(results.items)
        assert no_filter_results.total == results.total
        assert [item.score for item in no_filter_results.items] == [item.score for item in results.items]

    @pytest.mark.parametrize(
        "property, filter_match_value, filter_unmatch_value",
        [("value", "A", "C"), ("score", 0.5, 0), ("agent", "peter", "john"), ("type", "human", "model")],
    )
    async def test_search_with_suggestion_filter(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        property: str,
        filter_match_value: str,
        filter_unmatch_value: str,
    ):
        text_field = await TextFieldFactory.create()
        label_question = await QuestionFactory.create(
            settings={
                "type": QuestionType.label_selection.value,
                "options": [{"value": "A"}, {"value": "B"}],
            }
        )
        dataset = await DatasetFactory.create(fields=[text_field], questions=[label_question])
        records = await RecordFactory.create_batch(
            size=2,
            dataset=dataset,
            fields={text_field.name: f"This is the value for {text_field.name}"},
        )

        await SuggestionFactory.create(
            record=records[0], question=label_question, value="A", type="human", agent="peter", score=0.5
        )
        await SuggestionFactory.create(record=records[1], question=label_question, value="B")

        await refresh_dataset(dataset)
        await refresh_records(records)

        await search_engine.create_index(dataset)
        await search_engine.index_records(dataset, records)

        scope = SuggestionFilterScope(question=label_question.name, property=property)

        await self._check_filter_match(
            dataset, search_engine, filter=TermsFilter(scope=scope, values=[filter_match_value])
        )
        await self._check_filter_unmatch(
            dataset, search_engine, filter=TermsFilter(scope=scope, values=[filter_unmatch_value])
        )

    async def _check_filter_unmatch(
        self, dataset: Dataset, elasticsearch_engine: BaseElasticAndOpenSearchEngine, filter: TermsFilter
    ) -> None:
        result = await elasticsearch_engine.search(dataset, filter=filter)

        assert len(result.items) == 0
        assert result.total == 0

    async def _check_filter_match(
        self, dataset: Dataset, elasticsearch_engine: BaseElasticAndOpenSearchEngine, filter: TermsFilter
    ):
        result = await elasticsearch_engine.search(dataset, filter=filter)

        assert len(result.items) == 1
        assert result.total == 1

    @pytest.mark.parametrize(("offset", "limit"), [(0, 50), (10, 5), (0, 0), (90, 100)])
    async def test_search_with_pagination(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        dataset_for_pagination: Dataset,
        offset: int,
        limit: int,
    ):
        results = await search_engine.search(dataset_for_pagination, query="documents", offset=offset, limit=limit)

        assert len(results.items) == min(len(dataset_for_pagination.records) - offset, limit)
        assert results.total == 100

        records = sorted(dataset_for_pagination.records, key=lambda r: r.id)
        assert [record.id for record in records[offset : offset + limit]] == [item.record_id for item in results.items]

    @pytest.mark.parametrize(
        ("sort_by"),
        [
            SortBy(field="inserted_at"),
            SortBy(field="updated_at"),
            SortBy(field="inserted_at", order="desc"),
            SortBy(field="updated_at", order="desc"),
        ],
    )
    async def test_search_with_sort_by(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        test_banking_sentiment_dataset: Dataset,
        sort_by: SortBy,
    ):
        def _local_sort_by(record: Record) -> Any:
            if isinstance(sort_by.field, str):
                return getattr(record, sort_by.field)
            return record.metadata_[sort_by.field.name]

        results = await search_engine.search(test_banking_sentiment_dataset, sort_by=[sort_by])

        records = test_banking_sentiment_dataset.records
        if sort_by:
            records = sorted(records, key=_local_sort_by, reverse=sort_by.order == "desc")

        assert [item.record_id for item in results.items] == [record.id for record in records]

    async def test_index_records(self, search_engine: BaseElasticAndOpenSearchEngine, opensearch: OpenSearch):
        text_fields = await TextFieldFactory.create_batch(5)
        dataset = await DatasetFactory.create(fields=text_fields, questions=[])
        records = await RecordFactory.create_batch(
            size=10,
            dataset=dataset,
            fields={field.name: f"This is the value for {field.name}" for field in text_fields},
            responses=[],
        )

        await refresh_dataset(dataset)
        await refresh_records(records)

        await search_engine.create_index(dataset)
        await search_engine.index_records(dataset, records)

        index_name = es_index_name_for_dataset(dataset)

        es_docs = [hit["_source"] for hit in opensearch.search(index=index_name)["hits"]["hits"]]
        assert es_docs == [
            {
                "id": str(record.id),
                "fields": record.fields,
                "inserted_at": record.inserted_at.isoformat(),
                "updated_at": record.updated_at.isoformat(),
            }
            for record in records
        ]

    async def test_configure_metadata_property(
        self, search_engine: BaseElasticAndOpenSearchEngine, opensearch: OpenSearch
    ):
        dataset = await DatasetFactory.create()
        await refresh_dataset(dataset)

        await search_engine.create_index(dataset)

        terms_property = await TermsMetadataPropertyFactory.create(name="terms")
        await search_engine.configure_metadata_property(dataset, terms_property)

        index_name = es_index_name_for_dataset(dataset)
        index = opensearch.indices.get(index=index_name)[index_name]
        assert index["mappings"]["properties"]["metadata"] == {
            "dynamic": "false",
            "properties": {str(terms_property.name): {"type": "keyword"}},
        }

    async def test_index_records_with_suggestions(
        self, search_engine: BaseElasticAndOpenSearchEngine, opensearch: OpenSearch
    ):
        text_field = await TextFieldFactory.create()
        label_question = await QuestionFactory.create(
            settings={
                "type": QuestionType.label_selection.value,
                "options": [{"value": "A"}, {"value": "B"}],
            }
        )
        dataset = await DatasetFactory.create(fields=[text_field], questions=[label_question])
        records = await RecordFactory.create_batch(
            size=2,
            dataset=dataset,
            fields={text_field.name: f"This is the value for {text_field.name}"},
            responses=[],
        )

        await SuggestionFactory.create(record=records[0], question=label_question, value="A")
        await SuggestionFactory.create(record=records[1], question=label_question, value="B")

        await refresh_dataset(dataset)
        await refresh_records(records)

        await search_engine.create_index(dataset)
        await search_engine.index_records(dataset, records)

        index_name = es_index_name_for_dataset(dataset)

        es_docs = [hit["_source"] for hit in opensearch.search(index=index_name)["hits"]["hits"]]
        assert es_docs == [
            {
                "id": str(records[0].id),
                "fields": records[0].fields,
                "inserted_at": records[0].inserted_at.isoformat(),
                "updated_at": records[0].updated_at.isoformat(),
                "suggestions": {label_question.name: {"agent": None, "score": None, "type": None, "value": "A"}},
            },
            {
                "id": str(records[1].id),
                "fields": records[1].fields,
                "inserted_at": records[1].inserted_at.isoformat(),
                "updated_at": records[1].updated_at.isoformat(),
                "suggestions": {label_question.name: {"agent": None, "score": None, "type": None, "value": "B"}},
            },
        ]

    async def test_index_records_with_metadata(
        self, search_engine: BaseElasticAndOpenSearchEngine, opensearch: OpenSearch
    ):
        text_fields = await TextFieldFactory.create_batch(5)
        metadata_properties = await TermsMetadataPropertyFactory.create_batch(3)

        dataset = await DatasetFactory.create(fields=text_fields, metadata_properties=metadata_properties, questions=[])
        records = await RecordFactory.create_batch(
            size=10,
            dataset=dataset,
            fields={field.name: f"This is the value for {field.name}" for field in text_fields},
            metadata_={metadata_prop.name: "Value for Metadata Property" for metadata_prop in metadata_properties},
            responses=[],
        )

        await refresh_dataset(dataset)
        await refresh_records(records)

        await search_engine.create_index(dataset)
        await search_engine.index_records(dataset, records)

        index_name = es_index_name_for_dataset(dataset)

        es_docs = [hit["_source"] for hit in opensearch.search(index=index_name)["hits"]["hits"]]
        assert es_docs == [
            {
                "id": str(record.id),
                "fields": record.fields,
                "inserted_at": record.inserted_at.isoformat(),
                "updated_at": record.updated_at.isoformat(),
                "metadata": {
                    str(metadata_prop.name): record.metadata_[metadata_prop.name]
                    for metadata_prop in metadata_properties
                },
            }
            for record in records
        ]

    async def test_index_records_with_vectors(
        self, search_engine: BaseElasticAndOpenSearchEngine, opensearch: OpenSearch
    ):
        dataset = await DatasetFactory.create()
        text_fields = await TextFieldFactory.create_batch(size=5, dataset=dataset)
        vectors_settings = await VectorSettingsFactory.create_batch(size=5, dataset=dataset, dimensions=5)
        records = await RecordFactory.create_batch(
            size=5, fields={field.name: f"This is the value for {field.name}" for field in text_fields}, responses=[]
        )

        for record in records:
            for vector_settings in vectors_settings:
                await VectorFactory.create(
                    record=record, vector_settings=vector_settings, value=[1.0, 2.0, 3.0, 4.0, 5.0]
                )

        await refresh_dataset(dataset)
        await refresh_records(records)

        await search_engine.create_index(dataset)
        await search_engine.index_records(dataset, records)

        index_name = es_index_name_for_dataset(dataset)

        es_docs = [hit["_source"] for hit in opensearch.search(index=index_name)["hits"]["hits"]]
        assert es_docs == [
            {
                "id": str(record.id),
                "fields": record.fields,
                "inserted_at": record.inserted_at.isoformat(),
                "updated_at": record.updated_at.isoformat(),
                "vectors": {str(vector_settings.id): [1.0, 2.0, 3.0, 4.0, 5.0] for vector_settings in vectors_settings},
            }
            for record in records
        ]

    async def test_delete_records(self, search_engine: BaseElasticAndOpenSearchEngine, opensearch: OpenSearch):
        text_fields = await TextFieldFactory.create_batch(5)
        dataset = await DatasetFactory.create(fields=text_fields, questions=[])
        records = await RecordFactory.create_batch(
            size=10,
            dataset=dataset,
            fields={field.name: f"This is the value for {field.name}" for field in text_fields},
            responses=[],
        )

        await refresh_dataset(dataset)
        await refresh_records(records)

        await search_engine.create_index(dataset)
        await search_engine.index_records(dataset, records)

        records_to_delete, records_to_keep = records[:5], records[5:]
        await search_engine.delete_records(dataset, records_to_delete)

        index_name = es_index_name_for_dataset(dataset)

        deleted_docs = [
            hit["_source"]
            for hit in opensearch.search(
                index=index_name, body={"query": {"ids": {"values": [str(record.id) for record in records_to_delete]}}}
            )["hits"]["hits"]
        ]
        assert len(deleted_docs) == 0

        es_docs = [
            hit["_source"]
            for hit in opensearch.search(
                index=index_name, body={"query": {"ids": {"values": [str(record.id) for record in records_to_keep]}}}
            )["hits"]["hits"]
        ]
        assert len(records_to_keep) == 5

    async def test_update_record_response(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        test_banking_sentiment_dataset: Dataset,
    ):
        record = test_banking_sentiment_dataset.records[0]
        question = test_banking_sentiment_dataset.questions[0]

        response = await ResponseFactory.create(record=record, values={question.name: {"value": "test"}})
        record = await response.awaitable_attrs.record
        await record.awaitable_attrs.dataset
        await search_engine.update_record_response(response)

        index_name = es_index_name_for_dataset(test_banking_sentiment_dataset)

        results = opensearch.get(index=index_name, id=record.id)

        assert results["_source"]["responses"] == {
            response.user.username: {
                "values": {question.name: "test"},
                "status": response.status.value,
            }
        }

        index = opensearch.indices.get(index=index_name)[index_name]
        assert index["mappings"]["properties"]["responses"] == {
            "dynamic": "true",
            "properties": {
                response.user.username: {
                    "properties": {
                        "status": {"type": "keyword", "copy_to": [ALL_RESPONSES_STATUSES_FIELD]},
                        "values": {"properties": {question.name: {"index": False, "type": "text"}}},
                    }
                }
            },
        }

    async def test_delete_record_response(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        db: AsyncSession,
        test_banking_sentiment_dataset: Dataset,
    ):
        record = test_banking_sentiment_dataset.records[0]
        question = test_banking_sentiment_dataset.questions[0]

        response = await ResponseFactory.create(record=record, values={question.name: {"value": "test"}})
        record = await response.awaitable_attrs.record
        await record.awaitable_attrs.dataset
        await search_engine.update_record_response(response)

        index_name = es_index_name_for_dataset(test_banking_sentiment_dataset)

        results = opensearch.get(index=index_name, id=record.id)
        assert results["_source"]["responses"] == {
            response.user.username: {
                "values": {question.name: "test"},
                "status": response.status.value,
            }
        }

        await search_engine.delete_record_response(response)

        results = opensearch.get(index=index_name, id=record.id)
        assert results["_source"]["responses"] == {}

    @pytest.mark.parametrize(
        ("property_name", "expected_metrics"),
        [
            (
                "label",
                {
                    "total": 8,
                    "type": MetadataPropertyType.terms,
                    "values": [
                        {"count": 4, "term": "neutral"},
                        {"count": 3, "term": "negative"},
                        {"count": 1, "term": "positive"},
                    ],
                },
            ),
            ("textId", {"max": 8, "min": 0, "type": MetadataPropertyType.integer}),
            ("seq_float", {"max": 120020.1328125, "min": -149.1300048828125, "type": MetadataPropertyType.float}),
        ],
    )
    async def test_compute_metrics_for(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        db: "AsyncSession",
        test_banking_sentiment_dataset: Dataset,
        property_name: str,
        expected_metrics: dict,
    ):
        for property in test_banking_sentiment_dataset.metadata_properties:
            if property.name == property_name:
                break

        metrics = await search_engine.compute_metrics_for(property)

        assert metrics.type == property.type
        assert metrics.dict() == expected_metrics

    @pytest.mark.parametrize(
        ("property_type", "expected_metrics"),
        [
            (MetadataPropertyType.terms, {"total": 0, "values": []}),
            (MetadataPropertyType.integer, {"min": None, "max": None}),
            (MetadataPropertyType.float, {"min": None, "max": None}),
        ],
    )
    async def test_compute_metrics_for_missing_property(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        db: "AsyncSession",
        property_type: MetadataPropertyType,
        expected_metrics: dict,
    ):
        dataset = await DatasetFactory.create()

        await refresh_dataset(dataset)
        await search_engine.create_index(dataset)

        if property_type == MetadataPropertyType.terms:
            property = await TermsMetadataPropertyFactory.create(dataset=dataset)
        elif property_type == MetadataPropertyType.integer:
            property = await IntegerMetadataPropertyFactory.create(dataset=dataset)
        elif property_type == MetadataPropertyType.float:
            property = await FloatMetadataPropertyFactory.create(dataset=dataset)
        else:
            pytest.fail(f"Wrong property type {property_type}")

        metrics = await search_engine.compute_metrics_for(property)

        assert metrics.dict() == {"type": property_type, **expected_metrics}

    async def test_similarity_search_with_incomplete_inputs(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        test_banking_sentiment_dataset_with_vectors: Dataset,
    ):
        settings: VectorSettings = test_banking_sentiment_dataset_with_vectors.vectors_settings[0]
        with pytest.raises(
            expected_exception=ValueError,
            match="Must provide either vector value or record to compute the similarity search",
        ):
            await search_engine.similarity_search(
                dataset=test_banking_sentiment_dataset_with_vectors, vector_settings=settings
            )

    async def test_similarity_search_with_too_much_inputs(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        test_banking_sentiment_dataset_with_vectors: Dataset,
    ):
        settings: VectorSettings = test_banking_sentiment_dataset_with_vectors.vectors_settings[0]
        with pytest.raises(
            expected_exception=ValueError,
            match="Must provide either vector value or record to compute the similarity search",
        ):
            await search_engine.similarity_search(
                dataset=test_banking_sentiment_dataset_with_vectors,
                vector_settings=settings,
                value=[1, 2, 3],
                record=test_banking_sentiment_dataset_with_vectors.records[0],
            )

    async def test_similarity_search_by_vector_value(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        test_banking_sentiment_dataset_with_vectors: Dataset,
    ):
        selected_record = test_banking_sentiment_dataset_with_vectors.records[0]
        selected_vector = selected_record.vectors[0]

        responses = await search_engine.similarity_search(
            dataset=test_banking_sentiment_dataset_with_vectors,
            vector_settings=selected_vector.vector_settings,
            value=selected_vector.value,
            max_results=1,
        )

        assert responses.total == 1
        assert responses.items[0].record_id == selected_record.id

    async def test_similarity_search_by_vector_value_with_order(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        test_banking_sentiment_dataset_with_vectors: Dataset,
    ):
        selected_record: Record = test_banking_sentiment_dataset_with_vectors.records[0]
        vector_settings: VectorSettings = test_banking_sentiment_dataset_with_vectors.vectors_settings[0]

        responses = await search_engine.similarity_search(
            dataset=test_banking_sentiment_dataset_with_vectors,
            vector_settings=vector_settings,
            value=selected_record.vectors[0].value,
            order=SimilarityOrder.least_similar,
            max_results=1,
        )

        assert responses.total == 1
        assert responses.items[0].record_id != selected_record.id

    @pytest.mark.parametrize(
        "user_response_status_filter",
        [
            None,
            UserResponseStatusFilter(statuses=[ResponseStatusFilter.missing, ResponseStatusFilter.draft]),
        ],
    )
    async def test_similarity_search_by_record_and_user_response_filter(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        test_banking_sentiment_dataset_with_vectors: Dataset,
        user_response_status_filter: UserResponseStatusFilter,
    ):
        selected_record: Record = test_banking_sentiment_dataset_with_vectors.records[0]
        vector_settings: VectorSettings = test_banking_sentiment_dataset_with_vectors.vectors_settings[0]

        if user_response_status_filter:
            test_user = await UserFactory.create()
            user_response_status_filter.user = test_user

        responses = await search_engine.similarity_search(
            dataset=test_banking_sentiment_dataset_with_vectors,
            vector_settings=vector_settings,
            record=selected_record,
            max_results=1,
            user_response_status_filter=user_response_status_filter,
        )

        assert responses.total == 1
        assert responses.items[0].record_id != selected_record.id

    async def test_similarity_search_by_record(
        self,
        search_engine: BaseElasticAndOpenSearchEngine,
        opensearch: OpenSearch,
        test_banking_sentiment_dataset_with_vectors: Dataset,
    ):
        selected_record: Record = test_banking_sentiment_dataset_with_vectors.records[0]
        vector_settings: VectorSettings = test_banking_sentiment_dataset_with_vectors.vectors_settings[0]

        responses = await search_engine.similarity_search(
            dataset=test_banking_sentiment_dataset_with_vectors,
            vector_settings=vector_settings,
            record=selected_record,
            max_results=1,
        )

        assert responses.total == 1
        assert responses.items[0].record_id != selected_record.id

    async def _configure_record_responses(
        self,
        opensearch: OpenSearch,
        dataset: Dataset,
        response_status: List[ResponseStatusFilter],
        number_of_answered_records: int,
        user: Optional[User] = None,
    ):
        index_name = es_index_name_for_dataset(dataset)

        all_statuses = [
            ResponseStatusFilter.draft,
            ResponseStatusFilter.missing,
            ResponseStatusFilter.discarded,
            ResponseStatusFilter.submitted,
        ]
        selected_records = dataset.records[:number_of_answered_records]
        rest_of_the_records = dataset.records[number_of_answered_records:]

        # Create two responses with the same status (one in each record)
        for i, status in enumerate(response_status):
            if status != ResponseStatusFilter.missing:
                await self._update_records_responses(opensearch, index_name, selected_records, status, user)

        for status in all_statuses:
            if status not in response_status and status != ResponseStatusFilter.missing:
                await self._update_records_responses(opensearch, index_name, rest_of_the_records, status, user)

        opensearch.indices.refresh(index=index_name)

    async def _update_records_responses(
        self,
        opensearch: OpenSearch,
        index_name: str,
        records: List[Record],
        status: ResponseStatusFilter,
        user: Optional[User] = None,
    ):
        another_user = await UserFactory.create()

        for record in records:
            users_responses = {f"{another_user.username}.status": status.value}
            if user:
                users_responses.update({f"{user.username}.status": status.value})
            opensearch.update(index_name, id=record.id, body={"doc": {"responses": users_responses}})



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING
from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException, Query, Security, status
from fastapi import Response as HTTPResponse
from sqlalchemy.ext.asyncio import AsyncSession

from argilla_server.contexts import datasets
from argilla_server.database import get_async_db
from argilla_server.models import Record, User
from argilla_server.policies import RecordPolicyV1, authorize
from argilla_server.schemas.v1.records import Record as RecordSchema
from argilla_server.schemas.v1.records import RecordUpdate
from argilla_server.schemas.v1.responses import Response, ResponseCreate
from argilla_server.schemas.v1.suggestions import Suggestion, SuggestionCreate, Suggestions
from argilla_server.search_engine import SearchEngine, get_search_engine
from argilla_server.security import auth
from argilla_server.utils import parse_uuids

DELETE_RECORD_SUGGESTIONS_LIMIT = 100

router = APIRouter(tags=["records"])


async def _get_record(
    db: AsyncSession,
    record_id: UUID,
    with_dataset: bool = False,
    with_suggestions: bool = False,
    with_vectors: bool = False,
) -> Record:
    record = await datasets.get_record_by_id(db, record_id, with_dataset, with_suggestions, with_vectors)
    if not record:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Record with id `{record_id}` not found",
        )
    return record


@router.get("/records/{record_id}", response_model=RecordSchema)
async def get_record(
    *,
    db: AsyncSession = Depends(get_async_db),
    record_id: UUID,
    current_user: User = Security(auth.get_current_user),
):
    record = await _get_record(db, record_id, with_dataset=True, with_suggestions=True)

    await authorize(current_user, RecordPolicyV1.get(record))

    return record


@router.patch("/records/{record_id}", status_code=status.HTTP_200_OK, response_model=RecordSchema)
async def update_record(
    *,
    db: AsyncSession = Depends(get_async_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    record_id: UUID,
    record_update: RecordUpdate,
    current_user: User = Security(auth.get_current_user),
):
    record = await _get_record(db, record_id, with_dataset=True, with_suggestions=True, with_vectors=True)

    await authorize(current_user, RecordPolicyV1.update(record))

    try:
        return await datasets.update_record(db, search_engine, record, record_update)
    except ValueError as err:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(err))


@router.post("/records/{record_id}/responses", status_code=status.HTTP_201_CREATED, response_model=Response)
async def create_record_response(
    *,
    db: AsyncSession = Depends(get_async_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    record_id: UUID,
    response_create: ResponseCreate,
    current_user: User = Security(auth.get_current_user),
):
    record = await _get_record(db, record_id, with_dataset=True)

    await authorize(current_user, RecordPolicyV1.create_response(record))

    if await datasets.get_response_by_record_id_and_user_id(db, record_id, current_user.id):
        raise HTTPException(
            status_code=status.HTTP_409_CONFLICT,
            detail=f"Response already exists for record with id `{record_id}` and by user with id `{current_user.id}`",
        )

    # TODO: We should split API v1 into different FastAPI apps so we can customize error management.
    # After mapping ValueError to 422 errors for API v1 then we can remove this try except.
    try:
        return await datasets.create_response(db, search_engine, record, current_user, response_create)
    except ValueError as err:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(err))


@router.get("/records/{record_id}/suggestions", status_code=status.HTTP_200_OK, response_model=Suggestions)
async def get_record_suggestions(
    *, db: AsyncSession = Depends(get_async_db), record_id: UUID, current_user: User = Security(auth.get_current_user)
):
    record = await _get_record(db, record_id, with_dataset=True, with_suggestions=True)

    await authorize(current_user, RecordPolicyV1.get_suggestions(record))

    return Suggestions(items=record.suggestions)


@router.put(
    "/records/{record_id}/suggestions",
    summary="Create or update a suggestion",
    responses={
        status.HTTP_200_OK: {"model": Suggestion, "description": "Suggestion updated"},
        status.HTTP_201_CREATED: {"model": Suggestion, "description": "Suggestion created"},
    },
    status_code=status.HTTP_201_CREATED,
    response_model=Suggestion,
)
async def upsert_suggestion(
    *,
    db: AsyncSession = Depends(get_async_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    record_id: UUID,
    suggestion_create: SuggestionCreate,
    current_user: User = Security(auth.get_current_user),
    response: HTTPResponse,
):
    record = await _get_record(db, record_id, with_dataset=True)

    await authorize(current_user, RecordPolicyV1.create_suggestion(record))

    question = await datasets.get_question_by_id(db, suggestion_create.question_id)
    if not question:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Question with id `{suggestion_create.question_id}` not found",
        )

    if await datasets.get_suggestion_by_record_id_and_question_id(db, record_id, suggestion_create.question_id):
        # There is already a suggestion for this record and question, so we update it.
        response.status_code = status.HTTP_200_OK

    # TODO: We should split API v1 into different FastAPI apps so we can customize error management.
    # After mapping ValueError to 422 errors for API v1 then we can remove this try except.
    try:
        return await datasets.upsert_suggestion(db, search_engine, record, question, suggestion_create)
    except ValueError as err:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=str(err))


@router.delete(
    "/records/{record_id}/suggestions",
    summary="Delete suggestions for a record",
    status_code=status.HTTP_204_NO_CONTENT,
)
async def delete_record_suggestions(
    *,
    db: AsyncSession = Depends(get_async_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    record_id: UUID,
    current_user: User = Security(auth.get_current_user),
    ids: str = Query(..., description="A comma separated list with the IDs of the suggestions to be removed"),
):
    record = await _get_record(db, record_id, with_dataset=True)

    await authorize(current_user, RecordPolicyV1.delete_suggestions(record))

    suggestion_ids = parse_uuids(ids)
    num_suggestions = len(suggestion_ids)

    if num_suggestions == 0:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail="No suggestions IDs provided")

    if num_suggestions > DELETE_RECORD_SUGGESTIONS_LIMIT:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=f"Cannot delete more than {DELETE_RECORD_SUGGESTIONS_LIMIT} suggestions at once",
        )

    await datasets.delete_suggestions(db, search_engine, record, suggestion_ids)


@router.delete("/records/{record_id}", response_model=RecordSchema, response_model_exclude_unset=True)
async def delete_record(
    *,
    db: AsyncSession = Depends(get_async_db),
    search_engine: SearchEngine = Depends(get_search_engine),
    record_id: UUID,
    current_user: User = Security(auth.get_current_user),
):
    record = await _get_record(db, record_id, with_dataset=True)

    await authorize(current_user, RecordPolicyV1.delete(record))

    return await datasets.delete_record(db, search_engine, record)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING
from uuid import uuid4

import pytest
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.models import Suggestion, UserRole
from argilla_server.search_engine import SearchEngine
from sqlalchemy import func, select

from tests.factories import SuggestionFactory, UserFactory

if TYPE_CHECKING:
    from httpx import AsyncClient
    from sqlalchemy.ext.asyncio import AsyncSession


@pytest.mark.asyncio
class TestSuiteSuggestions:
    @pytest.mark.parametrize("role", [UserRole.admin, UserRole.owner])
    async def test_delete_suggestion(
        self, async_client: "AsyncClient", mock_search_engine: SearchEngine, db: "AsyncSession", role: UserRole
    ) -> None:
        suggestion = await SuggestionFactory.create()
        user = await UserFactory.create(role=role, workspaces=[suggestion.record.dataset.workspace])

        response = await async_client.delete(
            f"/api/v1/suggestions/{suggestion.id}",
            headers={API_KEY_HEADER_NAME: user.api_key},
        )

        assert response.status_code == 200
        assert response.json() == {
            "id": str(suggestion.id),
            "question_id": str(suggestion.question_id),
            "type": None,
            "score": None,
            "value": "negative",
            "agent": None,
        }
        assert (await db.execute(select(func.count(Suggestion.id)))).scalar() == 0

        mock_search_engine.delete_record_suggestion.assert_called_once_with(suggestion)

    async def test_delete_suggestion_non_existent(self, async_client: "AsyncClient", owner_auth_header: dict) -> None:
        response = await async_client.delete(f"/api/v1/suggestions/{uuid4()}", headers=owner_auth_header)

        assert response.status_code == 404

    async def test_delete_suggestion_as_admin_from_another_workspace(self, async_client: "AsyncClient") -> None:
        suggestion = await SuggestionFactory.create()
        user = await UserFactory.create(role=UserRole.admin)

        response = await async_client.delete(
            f"/api/v1/suggestions/{suggestion.id}", headers={API_KEY_HEADER_NAME: user.api_key}
        )

        assert response.status_code == 403

    async def test_delete_suggestion_as_annotator(self, async_client: "AsyncClient") -> None:
        suggestion = await SuggestionFactory.create()
        user = await UserFactory.create(role=UserRole.annotator, workspaces=[suggestion.record.dataset.workspace])

        response = await async_client.delete(
            f"/api/v1/suggestions/{suggestion.id}", headers={API_KEY_HEADER_NAME: user.api_key}
        )

        assert response.status_code == 403



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Any


def limit_value_length(data: Any, max_length: int) -> Any:
    """
    Given an input data, limits string values to a max_length by fetching
    last max_length characters

    Parameters
    ----------
    data:
        Input data
    max_length:
        Max length for string values

    Returns
    -------
        Limited version of data, if any
    """

    if isinstance(data, str):
        return data[-max_length:]
    if isinstance(data, dict):
        return {k: limit_value_length(v, max_length=max_length) for k, v in data.items()}
    if isinstance(data, (list, tuple, set)):
        new_values = map(lambda x: limit_value_length(x, max_length=max_length), data)
        return type(data)(new_values)
    return data



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, List, Literal, Optional, Union
from uuid import UUID

from argilla_server.models import SuggestionType
from argilla_server.pydantic_v1 import BaseModel, Field
from argilla_server.schemas.v1.questions import QuestionName

AGENT_REGEX = r"^(?=.*[a-zA-Z0-9])[a-zA-Z0-9-_:\.\/\s]+$"
AGENT_MIN_LENGTH = 1
AGENT_MAX_LENGTH = 200

SCORE_GREATER_THAN_OR_EQUAL = 0
SCORE_LESS_THAN_OR_EQUAL = 1


class SuggestionFilterScope(BaseModel):
    entity: Literal["suggestion"]
    question: QuestionName
    property: Optional[Union[Literal["value"], Literal["agent"], Literal["score"]]] = "value"


class SearchSuggestionOptionsQuestion(BaseModel):
    id: UUID
    name: str


class SearchSuggestionOptions(BaseModel):
    question: SearchSuggestionOptionsQuestion
    agents: List[str]


class SearchSuggestionsOptions(BaseModel):
    items: List[SearchSuggestionOptions]


class BaseSuggestion(BaseModel):
    question_id: UUID
    type: Optional[SuggestionType]
    value: Any
    agent: Optional[str]
    score: Optional[float]


class Suggestion(BaseSuggestion):
    id: UUID

    class Config:
        orm_mode = True


class Suggestions(BaseModel):
    items: List[Suggestion]


class SuggestionCreate(BaseSuggestion):
    agent: Optional[str] = Field(
        None,
        regex=AGENT_REGEX,
        min_length=AGENT_MIN_LENGTH,
        max_length=AGENT_MAX_LENGTH,
        description="Agent used to generate the suggestion",
    )
    score: Optional[float] = Field(
        None,
        ge=SCORE_GREATER_THAN_OR_EQUAL,
        le=SCORE_LESS_THAN_OR_EQUAL,
        description="The score assigned to the suggestion",
    )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List

from fastapi import Depends
from sqlalchemy.ext.asyncio import AsyncSession

from argilla_server.contexts import datasets
from argilla_server.database import get_async_db
from argilla_server.errors import future as errors
from argilla_server.models import User
from argilla_server.policies import RecordPolicyV1, authorize
from argilla_server.schemas.v1.responses import Response, ResponseBulk, ResponseBulkError, ResponseUpsert
from argilla_server.search_engine import SearchEngine, get_search_engine


class UpsertResponsesInBulkUseCase:
    def __init__(self, db: AsyncSession, search_engine: SearchEngine):
        self.db = db
        self.search_engine = search_engine

    async def execute(self, responses: List[ResponseUpsert], user: User) -> List[ResponseBulk]:
        responses_bulk_items = []

        all_records = await datasets.get_records_by_ids(self.db, [item.record_id for item in responses])
        non_empty_records = [r for r in all_records if r is not None]

        await datasets.preload_records_relationships_before_validate(self.db, non_empty_records)
        for item, record in zip(responses, all_records):
            try:
                if record is None:
                    raise errors.NotFoundError(f"Record with id `{item.record_id}` not found")

                await authorize(user, RecordPolicyV1.create_response(record))
                response = await datasets.upsert_response(self.db, self.search_engine, record, user, item)
            except Exception as err:
                responses_bulk_items.append(ResponseBulk(item=None, error=ResponseBulkError(detail=str(err))))
            else:
                responses_bulk_items.append(ResponseBulk(item=Response.from_orm(response), error=None))

        return responses_bulk_items


class UpsertResponsesInBulkUseCaseFactory:
    def __call__(
        self, db: AsyncSession = Depends(get_async_db), search_engine: SearchEngine = Depends(get_search_engine)
    ):
        return UpsertResponsesInBulkUseCase(db, search_engine)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING

import pytest
from argilla_server._version import __version__ as version
from argilla_server.services.info import ApiInfo, ApiStatus

if TYPE_CHECKING:
    from httpx import AsyncClient


@pytest.mark.asyncio
async def test_api_status(async_client: "AsyncClient"):
    response = await async_client.get("/api/_status")

    assert response.status_code == 200
    output = response.json()
    assert "version" in output and output["version"] == str(version)
    assert "elasticsearch" in output and isinstance(output["elasticsearch"], dict)
    assert "mem_info" in output and isinstance(output["mem_info"], dict)


@pytest.mark.asyncio
async def test_api_info(async_client: "AsyncClient"):
    response = await async_client.get("/api/_info")

    assert response.status_code == 200
    output = response.json()
    assert output["version"] == str(version)


@pytest.mark.asyncio
async def test_api_info(async_client: "AsyncClient"):
    response = await async_client.get("/api/_info")
    assert response.status_code == 200

    from argilla_server._version import __version__ as argilla_version

    info = ApiInfo.parse_obj(response.json())
    assert info.version == argilla_version


@pytest.mark.asyncio
async def test_api_status(async_client: "AsyncClient"):
    response = await async_client.get("/api/_status")
    assert response.status_code == 200

    from argilla_server._version import __version__ as argilla_version

    info = ApiStatus.parse_obj(response.json())
    assert info.version == argilla_version

    # Checking to not get the error dictionary service.py includes whenever something goes wrong
    assert "error" not in info.elasticsearch

    # Checking that the first key into mem_info dictionary has a nont-none value
    assert "rss" in info.mem_info is not None



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING, Type
from uuid import uuid4

import pytest
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.models import DatasetStatus, Question, UserRole
from argilla_server.schemas.v1.questions import QUESTION_CREATE_DESCRIPTION_MAX_LENGTH, QUESTION_CREATE_TITLE_MAX_LENGTH
from sqlalchemy import func, select

from tests.factories import (
    AnnotatorFactory,
    DatasetFactory,
    LabelSelectionQuestionFactory,
    MultiLabelSelectionQuestionFactory,
    RankingQuestionFactory,
    RatingQuestionFactory,
    TextQuestionFactory,
    UserFactory,
)

if TYPE_CHECKING:
    from httpx import AsyncClient
    from sqlalchemy.ext.asyncio import AsyncSession

    from tests.factories import QuestionFactory as QuestionFactoryType


@pytest.mark.parametrize(
    "QuestionFactory, payload, expected_settings",
    [
        (
            TextQuestionFactory,
            {
                "title": "New Title",
                "description": "New Description",
                "settings": {"type": "text", "use_markdown": True},
            },
            {"type": "text", "use_markdown": True},
        ),
        (
            TextQuestionFactory,
            {"description": None, "settings": {"type": "text"}},
            {"type": "text", "use_markdown": False},
        ),
        (
            TextQuestionFactory,
            {"name": "New Name", "required": True, "dataset_id": str(uuid4()), "settings": {"type": "text"}},
            {"type": "text", "use_markdown": False},
        ),
        (
            RatingQuestionFactory,
            {"name": "New Name", "description": "New Description", "settings": {"type": "rating"}},
            {
                "type": "rating",
                "options": [
                    {"value": 1},
                    {"value": 2},
                    {"value": 3},
                    {"value": 4},
                    {"value": 5},
                    {"value": 6},
                    {"value": 7},
                    {"value": 8},
                    {"value": 9},
                    {"value": 10},
                ],
            },
        ),
        (
            LabelSelectionQuestionFactory,
            {"settings": {"type": "label_selection", "visible_options": 3}},
            {
                "type": "label_selection",
                "options": [
                    {"value": "option1", "text": "Option 1", "description": None},
                    {"value": "option2", "text": "Option 2", "description": None},
                    {"value": "option3", "text": "Option 3", "description": None},
                ],
                "visible_options": 3,
            },
        ),
        (
            MultiLabelSelectionQuestionFactory,
            {"settings": {"type": "multi_label_selection", "visible_options": 3}},
            {
                "type": "multi_label_selection",
                "options": [
                    {"value": "option1", "text": "Option 1", "description": None},
                    {"value": "option2", "text": "Option 2", "description": None},
                    {"value": "option3", "text": "Option 3", "description": None},
                ],
                "visible_options": 3,
            },
        ),
        (
            LabelSelectionQuestionFactory,
            {"settings": {"type": "label_selection", "visible_options": None}},
            {
                "type": "label_selection",
                "options": [
                    {"value": "option1", "text": "Option 1", "description": None},
                    {"value": "option2", "text": "Option 2", "description": None},
                    {"value": "option3", "text": "Option 3", "description": None},
                ],
                "visible_options": None,
            },
        ),
        (
            RankingQuestionFactory,
            {"name": "New Name", "description": "New Description", "settings": {"type": "ranking"}},
            {
                "type": "ranking",
                "options": [
                    {"value": "completion-a", "text": "Completion A", "description": None},
                    {"value": "completion-b", "text": "Completion B", "description": None},
                    {"value": "completion-c", "text": "Completion C", "description": None},
                ],
            },
        ),
    ],
)
@pytest.mark.parametrize("role", [UserRole.owner])
@pytest.mark.asyncio
async def test_update_question(
    async_client: "AsyncClient",
    db: "AsyncSession",
    QuestionFactory: Type["QuestionFactoryType"],
    payload: dict,
    expected_settings: dict,
    role: UserRole,
):
    question = await QuestionFactory.create()
    user = await UserFactory.create(role=role, workspaces=[question.dataset.workspace])

    response = await async_client.patch(
        f"/api/v1/questions/{question.id}", headers={API_KEY_HEADER_NAME: user.api_key}, json=payload
    )

    title = payload.get("title") or question.title
    description = payload.get("description") or question.description

    assert response.status_code == 200, response.json()
    assert response.json() == {
        "id": str(question.id),
        "name": question.name,
        "title": title,
        "description": description,
        "required": False,
        "settings": expected_settings,
        "dataset_id": str(question.dataset_id),
        "inserted_at": question.inserted_at.isoformat(),
        "updated_at": question.updated_at.isoformat(),
    }

    question = await db.get(Question, question.id)
    assert question.title == title
    assert question.description == description
    assert question.settings == expected_settings


@pytest.mark.parametrize("title", [None, "", "t" * (QUESTION_CREATE_TITLE_MAX_LENGTH + 1)])
@pytest.mark.asyncio
async def test_update_question_with_invalid_title(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict, title: str
):
    question = await TextQuestionFactory.create(title="title")

    response = await async_client.patch(
        f"/api/v1/questions/{question.id}", headers=owner_auth_header, json={"title": title}
    )

    assert response.status_code == 422

    question = await db.get(Question, question.id)
    assert question.title == "title"


@pytest.mark.asyncio
async def test_update_question_with_description_as_none(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    question = await TextQuestionFactory.create(description="description")

    response = await async_client.patch(
        f"/api/v1/questions/{question.id}", headers=owner_auth_header, json={"description": None}
    )

    assert response.status_code == 200
    assert response.json()["description"] == None

    question = await db.get(Question, question.id)
    assert question.description == None


@pytest.mark.parametrize("description", ["", "d" * (QUESTION_CREATE_DESCRIPTION_MAX_LENGTH + 1)])
@pytest.mark.asyncio
async def test_update_question_with_invalid_description(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict, description: str
):
    question = await TextQuestionFactory.create(description="description")

    response = await async_client.patch(
        f"/api/v1/questions/{question.id}", headers=owner_auth_header, json={"description": description}
    )

    assert response.status_code == 422

    question = await db.get(Question, question.id)
    assert question.description == "description"


@pytest.mark.parametrize(
    "QuestionFactory, payload",
    [
        (TextQuestionFactory, {"settings": None}),
        (TextQuestionFactory, {"title": None, "description": None, "settings": None}),
        (TextQuestionFactory, {"settings": {"type": "text", "use_markdown": None}}),
        (TextQuestionFactory, {"title": "New Title", "settings": {"type": "label_selection"}}),
        (LabelSelectionQuestionFactory, {"settings": {"type": "label_selection", "visible_options": -5}}),
        (MultiLabelSelectionQuestionFactory, {"settings": {"type": "multi_label_selection", "visible_options": -5}}),
    ],
)
@pytest.mark.asyncio
async def test_update_question_with_invalid_settings(
    async_client: "AsyncClient", owner_auth_header: dict, QuestionFactory: Type["QuestionFactoryType"], payload: dict
):
    question = await QuestionFactory.create()

    response = await async_client.patch(f"/api/v1/questions/{question.id}", headers=owner_auth_header, json=payload)

    assert response.status_code == 422, payload


@pytest.mark.asyncio
async def test_update_question_with_invalid_payload(async_client: "AsyncClient", owner_auth_header: dict):
    question = await TextQuestionFactory.create()

    response = await async_client.patch(
        f"/api/v1/questions/{question.id}",
        headers=owner_auth_header,
        json={"title": {"this": "is", "not": "valid"}, "settings": {"use_markdown": "no"}},
    )

    assert response.status_code == 422


@pytest.mark.asyncio
async def test_update_question_non_existent(async_client: "AsyncClient", owner_auth_header: dict):
    response = await async_client.patch(
        f"/api/v1/questions/{uuid4()}",
        headers=owner_auth_header,
        json={"title": "New Title", "settings": {"type": "text", "use_markdown": True}},
    )

    assert response.status_code == 404


@pytest.mark.asyncio
async def test_update_question_as_admin_from_different_workspace(async_client: "AsyncClient"):
    question = await TextQuestionFactory.create()
    user = await UserFactory.create(role=UserRole.admin)

    response = await async_client.patch(
        f"/api/v1/questions/{question.id}",
        headers={API_KEY_HEADER_NAME: user.api_key},
        json={"title": "New Title", "settings": {"type": "text", "use_markdown": True}},
    )

    assert response.status_code == 403


@pytest.mark.asyncio
async def test_update_question_as_annotator(async_client: "AsyncClient"):
    question = await TextQuestionFactory.create()
    user = await AnnotatorFactory.create(workspaces=[question.dataset.workspace])

    response = await async_client.patch(
        f"/api/v1/questions/{question.id}",
        headers={API_KEY_HEADER_NAME: user.api_key},
        json={"title": "New Title", "settings": {"type": "text", "use_markdown": True}},
    )

    assert response.status_code == 403


@pytest.mark.parametrize("role", [UserRole.owner, UserRole.admin])
@pytest.mark.asyncio
async def test_delete_question(async_client: "AsyncClient", db: "AsyncSession", role: UserRole):
    question = await TextQuestionFactory.create(name="name", title="title", description="description")
    user = await UserFactory.create(role=role, workspaces=[question.dataset.workspace])

    response = await async_client.delete(
        f"/api/v1/questions/{question.id}", headers={API_KEY_HEADER_NAME: user.api_key}
    )

    assert response.status_code == 200
    assert (await db.execute(select(func.count(Question.id)))).scalar() == 0

    response_body = response.json()
    assert response_body == {
        "id": str(question.id),
        "name": "name",
        "title": "title",
        "description": "description",
        "required": False,
        "settings": {"type": "text", "use_markdown": False},
        "dataset_id": str(question.dataset_id),
        "inserted_at": question.inserted_at.isoformat(),
        "updated_at": question.updated_at.isoformat(),
    }


@pytest.mark.asyncio
async def test_delete_question_as_admin_from_different_workspace(async_client: "AsyncClient", db: "AsyncSession"):
    user = await UserFactory.create(role=UserRole.admin)
    question = await TextQuestionFactory.create()

    response = await async_client.delete(
        f"/api/v1/questions/{question.id}", headers={API_KEY_HEADER_NAME: user.api_key}
    )

    assert response.status_code == 403
    assert (await db.execute(select(func.count(Question.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_delete_question_without_authentication(async_client: "AsyncClient", db: "AsyncSession"):
    question = await TextQuestionFactory.create()

    response = await async_client.delete(f"/api/v1/questions/{question.id}")

    assert response.status_code == 401
    assert (await db.execute(select(func.count(Question.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_delete_question_as_annotator(async_client: "AsyncClient", db: "AsyncSession"):
    annotator = await AnnotatorFactory.create()
    question = await TextQuestionFactory.create()

    response = await async_client.delete(
        f"/api/v1/questions/{question.id}", headers={API_KEY_HEADER_NAME: annotator.api_key}
    )

    assert response.status_code == 403
    assert (await db.execute(select(func.count(Question.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_delete_question_belonging_to_published_dataset(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    dataset = await DatasetFactory.create(status=DatasetStatus.ready)
    question = await TextQuestionFactory.create(dataset=dataset)

    response = await async_client.delete(f"/api/v1/questions/{question.id}", headers=owner_auth_header)

    assert response.status_code == 422
    assert response.json() == {"detail": "Questions cannot be deleted for a published dataset"}
    assert (await db.execute(select(func.count(Question.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_delete_question_with_nonexistent_question_id(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    await TextQuestionFactory.create()

    response = await async_client.delete(f"/api/v1/questions/{uuid4()}", headers=owner_auth_header)

    assert response.status_code == 404
    assert (await db.execute(select(func.count(Question.id)))).scalar() == 1



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import functools
import importlib.metadata
import importlib.util
import operator
import re
import sys
from typing import Callable, Dict, List, Optional, TypeVar, Union

from packaging import version

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

ops = {
    "<": operator.lt,
    "<=": operator.le,
    "==": operator.eq,
    "!=": operator.ne,
    ">=": operator.ge,
    ">": operator.gt,
}


def _compare_versions(
    op: str,
    got_version: Optional[str],
    want_version: Optional[str],
    requirement: str,
    package: str,
    fn_name: Optional[str],
):
    if got_version is None or want_version is None:
        raise ValueError(
            f"Unable to compare versions for {requirement}: need={want_version} found={got_version}. This is unusual. Consider"
            f" reinstalling {package}."
        )
    if not ops[op](version.parse(got_version), version.parse(want_version)):
        raise ImportError(
            f"{requirement} must be installed{f' to use `{fn_name}`' if fn_name else ''}, but found {package}=={got_version}."
            f" You can install a supported version of '{package}' with this command: `pip install -U {requirement}`"
        )


def require_dependencies(requirements: Union[str, List[str]], fn_name: Optional[str] = None) -> None:
    """
    Perform a runtime check of the dependency versions, using the exact same syntax used by pip.
    The installed module version comes from the *site-packages* dir via *importlib.metadata*.

    Args:
        requirements: pip style definition of dependencies, e.g. "tokenizers==0.9.4", "tqdm>=4.27", "numpy". Can either be a single dependency or a list of dependencies.
        fn_name: what suggestion to print in case of requirements not being met

    Example:
    ```python
    require_dependencies("pandas>1.1.2")
    require_dependencies("datasets>1.17.0", "from_datasets")
    require_dependencies(["pandas>1.1.2", "datasets>1.17.0"])
    ```
    """

    if isinstance(requirements, list):
        if len(requirements) == 0:
            raise ValueError("requirements cannot be an empty list.")
        for requirement in requirements:
            require_dependencies(requirement, fn_name=fn_name)
        return

    # non-versioned check
    if re.match(r"^[\w_\-\d]+$", requirements):
        package, op, want_version = requirements, None, None
    else:
        match = re.findall(r"^([^!=<>\s]+)([\s!=<>]{1,2}.+)", requirements)
        if not match:
            raise ValueError(
                "requirement needs to be in the pip package format, .e.g., package_a==1.23, or package_b>=1.23, but"
                f" got {requirements!r}."
            )
        package, want_full = match[0]
        want_range = want_full.split(",")  # there could be multiple requirements
        wanted = {}
        for w in want_range:
            match = re.findall(r"^([\s!=<>]{1,2})(.+)", w)
            if not match:
                raise ValueError(
                    "requirement needs to be in the pip package format, .e.g., package_a==1.23, or package_b>=1.23,"
                    f" but got {requirements!r}."
                )
            op, want_version = match[0]
            wanted[op] = want_version
            if op not in ops:
                raise ValueError(f"{requirements}: need one of {list(ops.keys())}, but got {op!r}.")

    # special case
    if package == "python":
        got_version = ".".join([str(x) for x in sys.version_info[:3]])
        for op, want_version in wanted.items():
            _compare_versions(op, got_version, want_version, requirements, package, fn_name=fn_name)
        return

    # check if any version is installed
    try:
        got_version = importlib.metadata.version(package)
    except importlib.metadata.PackageNotFoundError:
        raise ModuleNotFoundError(
            f"'{package}' must be installed{f' to use `{fn_name}`' if fn_name else ''}! You can"
            f" install '{package}' with this command: `pip install {requirements}`"
        )

    # check that the right version is installed if version number or a range was provided
    if want_version is not None:
        for op, want_version in wanted.items():
            _compare_versions(op, got_version, want_version, requirements, package, fn_name=fn_name)


_P = ParamSpec("_P")
_R = TypeVar("_R")


def requires_dependencies(requirements: Union[str, List[str]]) -> Callable[[Callable[_P, _R]], Callable[_P, _R]]:
    """Decorator variant of `require_dependencies`.
    Perform a runtime check of the dependency versions, using the exact same syntax used by pip.
    The installed module version comes from the *site-packages* dir via *importlib.metadata*.

    Args:
        requirements: pip style definition, e.g.,  "tokenizers==0.9.4", "tqdm>=4.27", "numpy"

    Example:
    ```python
    @requires_dependencies("datasets>1.17.0")
    def from_datasets(self, ...):
        ...
    ```
    """

    def decorator(func: Callable[_P, _R]) -> Callable[_P, _R]:
        @functools.wraps(func)
        def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _R:
            require_dependencies(requirements, func.__name__)
            return func(*args, **kwargs)

        return wrapper

    return decorator


def _group_by_extra(dependencies: List[str]) -> Dict[str, List[str]]:
    grouped = {"base": []}

    for dep in dependencies:
        # Extract the name of the dependency by splitting at the first space or '['
        dep_name = dep.split()[0].split("[")[0]

        if "; extra ==" in dep:
            # Split the dependency and the extra value
            _, extra_value = dep.split(" ; extra ==")
            extra_value = extra_value.strip().strip("'")

            # Append the dependency name to the right extra group
            if extra_value not in grouped:
                grouped[extra_value] = []
            grouped[extra_value].append(dep_name)
        else:
            grouped["base"].append(dep_name)

    return grouped


def is_package_with_extras_installed(name: str, extras: List[str]) -> bool:
    """Checks the given extras of a package are installed.

    Args:
        name: the name of the package to check.
        extras: the extras to check.

    Returns:
        `True`, if the extras are installed, `False` otherwise.
    """
    try:
        requirements = importlib.metadata.requires(name)
    except importlib.metadata.PackageNotFoundError:
        return False

    grouped_requirements = _group_by_extra(requirements)
    available_extras = list(grouped_requirements.keys())
    for extra in extras:
        if extra not in available_extras:
            raise KeyError(f"'{name}' package does not provide '{extra}' extra")
        for requirement in grouped_requirements[extra]:
            try:
                importlib.metadata.version(requirement)
            except importlib.metadata.PackageNotFoundError:
                return False

    return True


__all__ = ["requires_dependencies", "require_dependencies", "is_package_with_extras_installed"]



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os
from datetime import datetime
from unittest.mock import call
from uuid import UUID, uuid4

import pytest
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.enums import ResponseStatus
from argilla_server.models import Response, User
from argilla_server.search_engine import SearchEngine
from argilla_server.use_cases.responses.upsert_responses_in_bulk import UpsertResponsesInBulkUseCase
from httpx import AsyncClient
from sqlalchemy import func, select
from sqlalchemy.ext.asyncio import AsyncSession

from tests.factories import (
    AnnotatorFactory,
    DatasetFactory,
    RatingQuestionFactory,
    RecordFactory,
    ResponseFactory,
    WorkspaceUserFactory,
)


@pytest.mark.asyncio
class TestCreateCurrentUserResponsesBulk:
    def url(self) -> str:
        return f"/api/v1/me/responses/bulk"

    def bulk_max_items(self) -> int:
        return 100

    async def test_multiple_responses(
        self, async_client: AsyncClient, db: AsyncSession, mock_search_engine: SearchEngine
    ):
        dataset = await DatasetFactory.create()
        await RatingQuestionFactory.create(name="prompt-quality", required=True, dataset=dataset)

        annotator = await AnnotatorFactory.create()
        await WorkspaceUserFactory.create(user_id=annotator.id, workspace_id=dataset.workspace.id)

        records = await RecordFactory.create_batch(3, dataset=dataset)

        other_dataset = await DatasetFactory.create()
        other_record = await RecordFactory.create(dataset=other_dataset)

        response_to_create_json = {
            "values": {"prompt-quality": {"value": 5}},
            "status": ResponseStatus.submitted,
            "record_id": str(records[0].id),
        }

        response_to_update = await ResponseFactory.create(
            status=ResponseStatus.draft,
            values={"prompt-quality": {"value": 1}},
            record=records[1],
            user=annotator,
        )

        response_to_update_json = {
            "values": {"prompt-quality": {"value": 10}},
            "status": ResponseStatus.submitted,
            "record_id": str(response_to_update.record_id),
        }

        invalid_response_json = {
            "values": {"non-existent-question": {"value": 10}},
            "status": ResponseStatus.submitted,
            "record_id": str(records[2].id),
        }

        unauthorized_response_json = {
            "status": ResponseStatus.draft,
            "record_id": str(other_record.id),
        }

        resp = await async_client.post(
            self.url(),
            headers={API_KEY_HEADER_NAME: annotator.api_key},
            json={
                "items": [
                    response_to_create_json,
                    response_to_update_json,
                    invalid_response_json,
                    unauthorized_response_json,
                ],
            },
        )

        assert resp.status_code == 200

        resp_json = resp.json()
        response_to_create_id = UUID(resp_json["items"][0]["item"]["id"])
        assert resp_json == {
            "items": [
                {
                    "item": {
                        "id": str(response_to_create_id),
                        "values": {"prompt-quality": {"value": 5}},
                        "status": ResponseStatus.submitted.value,
                        "record_id": str(records[0].id),
                        "user_id": str(annotator.id),
                        "inserted_at": datetime.fromisoformat(resp_json["items"][0]["item"]["inserted_at"]).isoformat(),
                        "updated_at": datetime.fromisoformat(resp_json["items"][0]["item"]["updated_at"]).isoformat(),
                    },
                    "error": None,
                },
                {
                    "item": {
                        "id": str(response_to_update.id),
                        "values": {"prompt-quality": {"value": 10}},
                        "status": ResponseStatus.submitted.value,
                        "record_id": str(records[1].id),
                        "user_id": str(annotator.id),
                        "inserted_at": datetime.fromisoformat(resp_json["items"][1]["item"]["inserted_at"]).isoformat(),
                        "updated_at": datetime.fromisoformat(resp_json["items"][1]["item"]["updated_at"]).isoformat(),
                    },
                    "error": None,
                },
                {
                    "item": None,
                    "error": {
                        "detail": "missing question with name=prompt-quality",
                    },
                },
                {
                    "item": None,
                    "error": {
                        "detail": "argilla.api.errors::ForbiddenOperationError(detail=Operation not allowed)",
                    },
                },
            ],
        }

        assert (await db.execute(select(func.count(Response.id)))).scalar() == 2

        response_to_create = (await db.execute(select(Response).filter_by(id=response_to_create_id))).scalar_one()
        await db.refresh(response_to_update)
        expected_calls = [
            call(response_to_create),
            call(response_to_update),
        ]
        mock_search_engine.update_record_response.assert_has_calls(expected_calls)

    async def test_response_to_create(
        self,
        async_client: AsyncClient,
        db: AsyncSession,
        mock_search_engine: SearchEngine,
        owner: User,
        owner_auth_header: dict,
    ):
        dataset = await DatasetFactory.create()
        await RatingQuestionFactory.create(name="prompt-quality", required=True, dataset=dataset)

        record = await RecordFactory.create(dataset=dataset)

        resp = await async_client.post(
            self.url(),
            headers=owner_auth_header,
            json={
                "items": [
                    {
                        "values": {"prompt-quality": {"value": 10}},
                        "status": ResponseStatus.submitted,
                        "record_id": str(record.id),
                    },
                ],
            },
        )

        assert resp.status_code == 200

        resp_json = resp.json()
        response_id = UUID(resp_json["items"][0]["item"]["id"])
        assert resp_json == {
            "items": [
                {
                    "item": {
                        "id": str(response_id),
                        "values": {"prompt-quality": {"value": 10}},
                        "status": ResponseStatus.submitted.value,
                        "record_id": str(record.id),
                        "user_id": str(owner.id),
                        "inserted_at": datetime.fromisoformat(resp_json["items"][0]["item"]["inserted_at"]).isoformat(),
                        "updated_at": datetime.fromisoformat(resp_json["items"][0]["item"]["updated_at"]).isoformat(),
                    },
                    "error": None,
                },
            ],
        }

        assert (await db.execute(select(func.count(Response.id)))).scalar() == 1

        response = (await db.execute(select(Response).filter_by(id=response_id))).scalar_one()
        mock_search_engine.update_record_response.assert_called_once_with(response)

    async def test_response_to_create_with_non_existent_record(
        self, async_client: AsyncClient, db: AsyncSession, mock_search_engine: SearchEngine, owner_auth_header: dict
    ):
        non_existent_record_id = uuid4()

        resp = await async_client.post(
            self.url(),
            headers=owner_auth_header,
            json={
                "items": [
                    {
                        "status": ResponseStatus.draft,
                        "record_id": str(non_existent_record_id),
                    },
                ],
            },
        )

        assert resp.status_code == 200
        assert resp.json() == {
            "items": [
                {
                    "item": None,
                    "error": {
                        "detail": f"Record with id `{non_existent_record_id}` not found",
                    },
                },
            ],
        }

        assert (await db.execute(select(func.count(Response.id)))).scalar() == 0
        assert not mock_search_engine.update_record_response.called

    async def test_response_to_update(
        self,
        async_client: AsyncClient,
        db: AsyncSession,
        mock_search_engine: SearchEngine,
        owner: User,
        owner_auth_header: dict,
    ):
        dataset = await DatasetFactory.create()
        await RatingQuestionFactory.create(name="prompt-quality", required=True, dataset=dataset)

        record = await RecordFactory.create(dataset=dataset)
        response = await ResponseFactory.create(
            status=ResponseStatus.draft,
            values={"prompt-quality": {"value": 1}},
            record=record,
            user=owner,
        )

        resp = await async_client.post(
            self.url(),
            headers=owner_auth_header,
            json={
                "items": [
                    {
                        "values": {"prompt-quality": {"value": 10}},
                        "status": ResponseStatus.submitted,
                        "record_id": str(record.id),
                    },
                ],
            },
        )

        assert resp.status_code == 200

        resp_json = resp.json()
        assert resp_json == {
            "items": [
                {
                    "item": {
                        "id": str(response.id),
                        "values": {"prompt-quality": {"value": 10}},
                        "status": ResponseStatus.submitted.value,
                        "record_id": str(record.id),
                        "user_id": str(owner.id),
                        "inserted_at": datetime.fromisoformat(resp_json["items"][0]["item"]["inserted_at"]).isoformat(),
                        "updated_at": datetime.fromisoformat(resp_json["items"][0]["item"]["updated_at"]).isoformat(),
                    },
                    "error": None,
                },
            ],
        }

        assert (await db.execute(select(func.count(Response.id)))).scalar() == 1

        await db.refresh(response)
        mock_search_engine.update_record_response.assert_called_once_with(response)

    async def test_invalid_response(
        self, async_client: AsyncClient, db: AsyncSession, mock_search_engine: SearchEngine, owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()
        record = await RecordFactory.create(dataset=dataset)

        resp = await async_client.post(
            self.url(),
            headers=owner_auth_header,
            json={
                "items": [
                    {
                        "values": {"prompt-quality": {"value": 10}},
                        "status": ResponseStatus.submitted,
                        "record_id": str(record.id),
                    },
                ],
            },
        )

        assert resp.status_code == 200
        assert resp.json() == {
            "items": [
                {
                    "item": None,
                    "error": {
                        "detail": "found responses for non configured questions: ['prompt-quality']",
                    },
                },
            ],
        }

        assert (await db.execute(select(func.count(Response.id)))).scalar() == 0
        assert not mock_search_engine.update_record_response.called

    async def test_unauthorized_response(
        self, async_client: AsyncClient, mock_search_engine: SearchEngine, db: AsyncSession
    ):
        dataset = await DatasetFactory.create()
        record = await RecordFactory.create(dataset=dataset)
        annotator = await AnnotatorFactory.create()

        resp = await async_client.post(
            self.url(),
            headers={API_KEY_HEADER_NAME: annotator.api_key},
            json={
                "items": [
                    {
                        "status": ResponseStatus.draft,
                        "record_id": str(record.id),
                    },
                ],
            },
        )

        assert resp.status_code == 200
        assert resp.json() == {
            "items": [
                {
                    "item": None,
                    "error": {
                        "detail": "argilla.api.errors::ForbiddenOperationError(detail=Operation not allowed)",
                    },
                },
            ],
        }

        assert (await db.execute(select(func.count(Response.id)))).scalar() == 0
        assert not mock_search_engine.update_record_response.called

    async def test_no_responses(self, async_client: AsyncClient, owner_auth_header: dict):
        resp = await async_client.post(
            self.url(),
            headers=owner_auth_header,
            json={"items": []},
        )

        assert resp.status_code == 422

    async def test_too_many_responses(self, async_client: AsyncClient, owner_auth_header: dict):
        resp = await async_client.post(
            self.url(),
            headers=owner_auth_header,
            json={"items": [{}] * (self.bulk_max_items() + 1)},
        )

        assert resp.status_code == 422

    @pytest.mark.skipif(reason="Profiling is not active", condition=not bool(os.getenv("TEST_PROFILING", None)))
    async def test_create_responses_in_bulk_profiling(self, db: "AsyncSession", elasticsearch_config: dict):
        from argilla_server.schemas.v1.responses import DraftResponseUpsert
        from argilla_server.search_engine import ElasticSearchEngine
        from pyinstrument import Profiler

        from tests.factories import OwnerFactory, TextFieldFactory

        async def refresh_dataset(dataset):
            await dataset.awaitable_attrs.fields
            await dataset.awaitable_attrs.questions
            await dataset.awaitable_attrs.metadata_properties
            await dataset.awaitable_attrs.vectors_settings

        async def refresh_records(records):
            for record in records:
                await record.awaitable_attrs.suggestions
                await record.awaitable_attrs.responses
                await record.awaitable_attrs.vectors

        dataset = await DatasetFactory.create()
        user = await OwnerFactory.create()

        await RatingQuestionFactory.create(name="prompt-quality", required=True, dataset=dataset)
        await TextFieldFactory.create(name="text", required=True, dataset=dataset)
        await TextFieldFactory.create(name="sentiment", required=True, dataset=dataset)

        records = await RecordFactory.create_batch(dataset=dataset, size=500)

        engine = ElasticSearchEngine(config=elasticsearch_config, number_of_replicas=0, number_of_shards=1)

        await refresh_dataset(dataset)
        await refresh_records(records)

        await engine.create_index(dataset)
        await engine.index_records(dataset, records)

        profiler = Profiler()

        responses = [
            DraftResponseUpsert.parse_obj(
                {
                    "values": {"prompt-quality": {"value": 10}},
                    "record_id": record.id,
                    "status": "draft",
                }
            )
            for record in records
        ]
        use_case = UpsertResponsesInBulkUseCase(db, engine)
        with profiler:
            bulk_items = await use_case.execute(responses, user)
            await use_case.execute([bulk_item.item for bulk_item in bulk_items], user)

        profiler.open_in_browser()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import uuid
from typing import TYPE_CHECKING, Type
from uuid import uuid4

import pytest
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.enums import MetadataPropertyType, UserRole
from argilla_server.models import MetadataProperty, UserRole
from argilla_server.schemas.v1.metadata_properties import METADATA_PROPERTY_CREATE_TITLE_MAX_LENGTH
from argilla_server.search_engine import FloatMetadataMetrics, IntegerMetadataMetrics, TermsMetadataMetrics
from sqlalchemy import func, select

from tests.factories import (
    AdminFactory,
    AnnotatorFactory,
    BaseFactory,
    FloatMetadataPropertyFactory,
    IntegerMetadataPropertyFactory,
    TermsMetadataPropertyFactory,
    UserFactory,
    WorkspaceUserFactory,
)

if TYPE_CHECKING:
    from argilla_server.search_engine import MetadataMetrics, SearchEngine
    from httpx import AsyncClient


@pytest.mark.asyncio
class TestSuiteMetadataProperties:
    @pytest.mark.parametrize(
        ("factory_class", "expected_metric", "expected_json"),
        [
            (
                TermsMetadataPropertyFactory,
                TermsMetadataMetrics(total=10, values=[TermsMetadataMetrics.TermCount(term="term", count=10)]),
                {"type": "terms", "total": 10, "values": [{"term": "term", "count": 10}]},
            ),
            (
                IntegerMetadataPropertyFactory,
                IntegerMetadataMetrics(min=10, max=100),
                {"type": "integer", "min": 10, "max": 100},
            ),
            (
                FloatMetadataPropertyFactory,
                FloatMetadataMetrics(min=10.3, max=11.32),
                {"type": "float", "min": 10.3, "max": 11.32},
            ),
            (
                FloatMetadataPropertyFactory,
                FloatMetadataMetrics(min=1.23456789, max=2.34567890),
                {"type": "float", "min": 1.23457, "max": 2.34568},
            ),
        ],
    )
    async def test_get_metadata_property_metrics(
        self,
        async_client: "AsyncClient",
        mock_search_engine: "SearchEngine",
        owner_auth_header: dict,
        factory_class: Type[BaseFactory],
        expected_metric: "MetadataMetrics",
        expected_json: dict,
    ):
        metadata_property = await factory_class.create()

        mock_search_engine.compute_metrics_for.return_value = expected_metric

        response = await async_client.get(
            f"/api/v1/metadata-properties/{metadata_property.id}/metrics", headers=owner_auth_header
        )

        assert response.status_code == 200
        assert response.json() == expected_json

    async def test_get_metadata_property_metrics_without_authentication(self, async_client: "AsyncClient"):
        metadata_property = await TermsMetadataPropertyFactory.create()

        response = await async_client.get(f"/api/v1/metadata-properties/{metadata_property.id}/metrics")

        assert response.status_code == 401

    @pytest.mark.parametrize("role", [UserRole.admin, UserRole.annotator])
    async def test_get_metadata_property_metrics_as_allowed_role(
        self, async_client: "AsyncClient", mock_search_engine: "SearchEngine", role: UserRole
    ):
        metadata_property = await IntegerMetadataPropertyFactory.create(allowed_roles=[role])
        workspace = metadata_property.dataset.workspace

        user = await UserFactory.create(role=role)

        await WorkspaceUserFactory.create(user_id=user.id, workspace_id=workspace.id)

        mock_search_engine.compute_metrics_for.return_value = IntegerMetadataMetrics(min=0, max=10)

        response = await async_client.get(
            f"/api/v1/metadata-properties/{metadata_property.id}/metrics", headers={API_KEY_HEADER_NAME: user.api_key}
        )

        assert response.status_code == 200

    @pytest.mark.parametrize("role", [UserRole.admin, UserRole.annotator])
    async def test_get_metadata_property_metrics_as_non_allowed_role(
        self, async_client: "AsyncClient", mock_search_engine: "SearchEngine", role: UserRole
    ):
        metadata_property = await IntegerMetadataPropertyFactory.create(allowed_roles=[])
        workspace = metadata_property.dataset.workspace

        user = await UserFactory.create(role=role)

        await WorkspaceUserFactory.create(user_id=user.id, workspace_id=workspace.id)

        mock_search_engine.compute_metrics_for.return_value = IntegerMetadataMetrics(min=0, max=10)

        response = await async_client.get(
            f"/api/v1/metadata-properties/{metadata_property.id}/metrics", headers={API_KEY_HEADER_NAME: user.api_key}
        )

        assert response.status_code == 403

    async def test_get_metadata_property_metrics_with_nonexistent_metadata_property(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        await TermsMetadataPropertyFactory.create()

        response = await async_client.get(
            f"/api/v1/metadata-properties/{uuid.uuid4()}/metrics", headers=owner_auth_header
        )

        assert response.status_code == 404

    @pytest.mark.parametrize("role", [UserRole.admin, UserRole.annotator])
    async def test_get_metadata_property_metrics_as_restricted_user_role_from_different_workspace(
        self, async_client: "AsyncClient", role: UserRole
    ):
        user = await UserFactory.create(role=role)
        metadata_property = await TermsMetadataPropertyFactory.create()

        response = await async_client.get(
            f"/api/v1/metadata-properties/{metadata_property.id}/metrics", headers={API_KEY_HEADER_NAME: user.api_key}
        )
        assert response.status_code == 403


@pytest.mark.asyncio
async def test_update_metadata_property(async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict):
    metadata_property = await IntegerMetadataPropertyFactory.create(
        name="name", title="title", allowed_roles=[UserRole.admin, UserRole.annotator]
    )

    assert metadata_property.visible_for_annotators == True

    response = await async_client.patch(
        f"/api/v1/metadata-properties/{metadata_property.id}",
        headers=owner_auth_header,
        json={"title": "updated title", "visible_for_annotators": False},
    )

    assert response.status_code == 200
    assert response.json() == {
        "id": str(metadata_property.id),
        "name": "name",
        "title": "updated title",
        "settings": {"type": "integer", "min": None, "max": None},
        "visible_for_annotators": False,
        "dataset_id": str(metadata_property.dataset.id),
        "inserted_at": metadata_property.inserted_at.isoformat(),
        "updated_at": metadata_property.updated_at.isoformat(),
    }

    metadata_property = await db.get(MetadataProperty, metadata_property.id)
    assert metadata_property.title == "updated title"
    assert metadata_property.visible_for_annotators == False
    assert metadata_property.allowed_roles == [UserRole.admin]


@pytest.mark.asyncio
async def test_update_metadata_property_without_authentication(async_client: "AsyncClient", db: "AsyncSession"):
    metadata_property = await IntegerMetadataPropertyFactory.create()

    response = await async_client.patch(f"/api/v1/metadata-properties/{metadata_property.id}")

    assert response.status_code == 401


@pytest.mark.asyncio
async def test_update_metadata_property_title(async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict):
    metadata_property = await IntegerMetadataPropertyFactory.create(title="title")

    response = await async_client.patch(
        f"/api/v1/metadata-properties/{metadata_property.id}",
        headers=owner_auth_header,
        json={"title": "updated title"},
    )

    assert response.status_code == 200

    response_json = response.json()
    assert response_json["title"] == "updated title"

    metadata_property = await db.get(MetadataProperty, metadata_property.id)
    assert metadata_property.title == "updated title"


@pytest.mark.parametrize(
    "title",
    [None, "", "t" * (METADATA_PROPERTY_CREATE_TITLE_MAX_LENGTH + 1)],
)
@pytest.mark.asyncio
async def test_update_metadata_property_with_invalid_title(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict, title: str
):
    metadata_property = await IntegerMetadataPropertyFactory.create(title="title")

    response = await async_client.patch(
        f"/api/v1/metadata-properties/{metadata_property.id}",
        headers=owner_auth_header,
        json={"title": title},
    )

    assert response.status_code == 422

    metadata_property = await db.get(MetadataProperty, metadata_property.id)
    assert metadata_property.title == "title"


@pytest.mark.asyncio
async def test_update_metadata_property_enabling_visible_for_annotators(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    metadata_property = await IntegerMetadataPropertyFactory.create(allowed_roles=[UserRole.admin])

    assert metadata_property.visible_for_annotators == False

    response = await async_client.patch(
        f"/api/v1/metadata-properties/{metadata_property.id}",
        headers=owner_auth_header,
        json={"visible_for_annotators": True},
    )

    assert response.status_code == 200

    response_json = response.json()
    assert response_json["visible_for_annotators"] == True

    metadata_property = await db.get(MetadataProperty, metadata_property.id)
    assert metadata_property.visible_for_annotators == True
    assert metadata_property.allowed_roles == [UserRole.admin, UserRole.annotator]


@pytest.mark.asyncio
async def test_update_metadata_property_disabling_visible_for_annotators(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    metadata_property = await IntegerMetadataPropertyFactory.create(allowed_roles=[UserRole.admin, UserRole.annotator])

    assert metadata_property.visible_for_annotators == True

    response = await async_client.patch(
        f"/api/v1/metadata-properties/{metadata_property.id}",
        headers=owner_auth_header,
        json={"visible_for_annotators": False},
    )

    assert response.status_code == 200

    response_json = response.json()
    assert response_json["visible_for_annotators"] == False

    metadata_property = await db.get(MetadataProperty, metadata_property.id)
    assert metadata_property.visible_for_annotators == False
    assert metadata_property.allowed_roles == [UserRole.admin]


@pytest.mark.asyncio
async def test_update_metadata_property_with_visible_for_annotators_as_none(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    metadata_property = await IntegerMetadataPropertyFactory.create(allowed_roles=[UserRole.admin, UserRole.annotator])

    assert metadata_property.visible_for_annotators == True

    response = await async_client.patch(
        f"/api/v1/metadata-properties/{metadata_property.id}",
        headers=owner_auth_header,
        json={"visible_for_annotators": None},
    )

    assert response.status_code == 422

    metadata_property = await db.get(MetadataProperty, metadata_property.id)
    assert metadata_property.visible_for_annotators == True
    assert metadata_property.allowed_roles == [UserRole.admin, UserRole.annotator]


@pytest.mark.asyncio
async def test_update_metadata_property_as_admin(async_client: "AsyncClient", db: "AsyncSession"):
    metadata_property = await IntegerMetadataPropertyFactory.create()
    admin = await AdminFactory.create(workspaces=[metadata_property.dataset.workspace])

    response = await async_client.patch(
        f"/api/v1/metadata-properties/{metadata_property.id}",
        headers={API_KEY_HEADER_NAME: admin.api_key},
        json={},
    )

    assert response.status_code == 200


@pytest.mark.asyncio
async def test_update_metadata_property_as_admin_from_different_workspace(
    async_client: "AsyncClient", db: "AsyncSession"
):
    metadata_property = await IntegerMetadataPropertyFactory.create()
    admin = await AdminFactory.create()

    response = await async_client.patch(
        f"/api/v1/metadata-properties/{metadata_property.id}",
        headers={API_KEY_HEADER_NAME: admin.api_key},
        json={},
    )

    assert response.status_code == 403


@pytest.mark.asyncio
async def test_update_metadata_property_as_annotator(async_client: "AsyncClient", db: "AsyncSession"):
    metadata_property = await IntegerMetadataPropertyFactory.create()
    annotator = await AnnotatorFactory.create(workspaces=[metadata_property.dataset.workspace])

    response = await async_client.patch(
        f"/api/v1/metadata-properties/{metadata_property.id}",
        headers={API_KEY_HEADER_NAME: annotator.api_key},
        json={},
    )

    assert response.status_code == 403


@pytest.mark.asyncio
async def test_update_metadata_property_with_empty_payload(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    metadata_property = await IntegerMetadataPropertyFactory.create()

    response = await async_client.patch(
        f"/api/v1/metadata-properties/{metadata_property.id}",
        headers=owner_auth_header,
        json={},
    )

    assert response.status_code == 200


@pytest.mark.asyncio
async def test_update_metadata_property_with_nonexistent_metadata_property_id(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    await IntegerMetadataPropertyFactory.create()

    response = await async_client.patch(
        f"/api/v1/metadata-properties/{uuid4()}",
        headers=owner_auth_header,
        json={},
    )

    assert response.status_code == 404


@pytest.mark.parametrize("user_role", [UserRole.owner, UserRole.admin])
@pytest.mark.asyncio
async def test_delete_metadata_property(async_client: "AsyncClient", db: "AsyncSession", user_role: UserRole):
    metadata_property = await IntegerMetadataPropertyFactory.create(name="name", title="title")

    user = await UserFactory.create(role=user_role, workspaces=[metadata_property.dataset.workspace])

    response = await async_client.delete(
        f"/api/v1/metadata-properties/{metadata_property.id}", headers={API_KEY_HEADER_NAME: user.api_key}
    )

    assert response.status_code == 200
    assert (await db.execute(select(func.count(MetadataProperty.id)))).scalar() == 0

    response_body = response.json()
    assert response_body == {
        "id": str(metadata_property.id),
        "name": "name",
        "title": "title",
        "settings": {"type": MetadataPropertyType.integer, "min": None, "max": None},
        "visible_for_annotators": True,
        "dataset_id": str(metadata_property.dataset_id),
        "inserted_at": metadata_property.inserted_at.isoformat(),
        "updated_at": metadata_property.updated_at.isoformat(),
    }


@pytest.mark.asyncio
async def test_delete_metadata_property_without_authentication(async_client: "AsyncClient", db: "AsyncSession"):
    metadata_property = await IntegerMetadataPropertyFactory.create()

    response = await async_client.delete(f"/api/v1/metadata-properties/{metadata_property.id}")

    assert response.status_code == 401
    assert (await db.execute(select(func.count(MetadataProperty.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_delete_metadata_property_as_admin_from_different_workspace(
    async_client: "AsyncClient", db: "AsyncSession"
):
    admin = await UserFactory.create(role=UserRole.admin)
    metadata_property = await IntegerMetadataPropertyFactory.create()

    response = await async_client.delete(
        f"/api/v1/metadata-properties/{metadata_property.id}", headers={API_KEY_HEADER_NAME: admin.api_key}
    )

    assert response.status_code == 403
    assert (await db.execute(select(func.count(MetadataProperty.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_delete_metadata_property_as_annotator(async_client: "AsyncClient", db: "AsyncSession"):
    annotator = await AnnotatorFactory.create()
    metadata_property = await IntegerMetadataPropertyFactory.create()

    response = await async_client.delete(
        f"/api/v1/metadata-properties/{metadata_property.id}", headers={API_KEY_HEADER_NAME: annotator.api_key}
    )

    assert response.status_code == 403
    assert (await db.execute(select(func.count(MetadataProperty.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_delete_metadata_property_with_nonexistent_metadata_property_id(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    await IntegerMetadataPropertyFactory.create()

    response = await async_client.delete(f"/api/v1/metadata-properties/{uuid4()}", headers=owner_auth_header)

    assert response.status_code == 404
    assert (await db.execute(select(func.count(MetadataProperty.id)))).scalar() == 1
